import logging
import sys

import numpy as np
import torch
from lab.torch import B
from stheno.torch import Graph, GP, EQ, RQ, Delta, Linear, ZeroKernel
from varz import Vars
from varz.torch import minimise_l_bfgs_b

from .model import GPAR, per_output

__all__ = ['GPARRegressor', 'log_transform', 'squishing_transform']

log = logging.getLogger(__name__)

#: Log transform for the data.
log_transform = (B.log, B.exp)

#: Squishing transform for the data.
squishing_transform = (lambda x: B.sign(x) * B.log(1 + B.abs(x)),
                       lambda x: B.sign(x) * (B.exp(B.abs(x)) - 1))


def _vector_from_init(init, length):
    # If only a single value is given, create ones.
    if np.size(init) == 1:
        return init * np.ones(length)

    # Multiple values are given. Check that enough values are available.
    init_squeezed = np.squeeze(init)
    if np.ndim(init_squeezed) != 1:
        raise ValueError('Incorrect shape {} of hyperparameters.'
                         ''.format(np.shape(init)))
    if np.size(init_squeezed) < length:  # Squeezed doesn't change size.
        raise ValueError('Not enough hyperparameters specified.')

    # Return initialisation.
    return np.array(init_squeezed)[:length]


def _determine_indices(m, pi, markov):
    # Build in the Markov structure: juggle with the indices of the outputs.
    p_last = pi - 1  # Index of last output that is given as input.
    p_start = 0 if markov is None else max(p_last - (markov - 1), 0)
    p_num = p_last - p_start + 1

    # Determine the indices corresponding to the outputs and inputs.
    m_inds = list(range(m))
    p_inds = list(range(m + p_start, m + p_last + 1))

    return m_inds, p_inds, p_num


def _model_generator(vs,
                     m,  # This is the _number_ of inputs.
                     pi,  # This is the _index_ of the output modelled.
                     scale,
                     scale_tie,
                     per,
                     per_period,
                     per_scale,
                     per_decay,
                     input_linear,
                     input_linear_scale,
                     linear,
                     linear_scale,
                     nonlinear,
                     nonlinear_scale,
                     rq,
                     markov,
                     noise):
    def model():
        # Start with a zero kernels.
        kernel_inputs = ZeroKernel()  # Kernel over inputs.
        kernel_outputs = ZeroKernel()  # Kernel over outputs.

        # Determine indices corresponding to the inputs and outputs.
        m_inds, p_inds, p_num = _determine_indices(m, pi, markov)

        # Add nonlinear kernel over the inputs.
        variance = vs.bnd(name='{}/input/var'.format(pi), init=1.)
        scales = vs.bnd(name='{}/input/scales'.format(0 if scale_tie else pi),
                        init=_vector_from_init(scale, m))
        if rq:
            k = RQ(vs.bnd(name='{}/input/alpha', init=1e-2,
                          lower=1e-3, upper=1e3))
        else:
            k = EQ()
        kernel_inputs += variance * k.stretch(scales)

        # Add a locally periodic kernel over the inputs.
        if per:
            variance = vs.bnd(name='{}/input/per/var'.format(pi), init=1.)
            scales = vs.bnd(name='{}/input/per/scales'.format(pi),
                            init=_vector_from_init(per_scale, 2 * m))
            periods = vs.bnd(name='{}/input/per/pers'.format(pi),
                             init=_vector_from_init(per_period, m))
            decays = vs.bnd(name='{}/input/per/decay'.format(pi),
                            init=_vector_from_init(per_decay, m))
            kernel_inputs += variance * \
                             EQ().stretch(scales).periodic(periods) * \
                             EQ().stretch(decays)

        # Add a linear kernel over the inputs.
        if input_linear:
            scales = vs.bnd(name='{}/input/lin/scales'.format(pi),
                            init=_vector_from_init(input_linear_scale, m))
            const = vs.get(name='{}/input/lin/const'.format(pi), init=1.)
            kernel_inputs += Linear().stretch(scales) + const

        # Add linear kernel over the outputs.
        if linear and pi > 0:
            scales = vs.bnd(name='{}/output/lin/scales'.format(pi),
                            init=_vector_from_init(linear_scale, p_num))
            kernel_outputs += Linear().stretch(scales)

        # Add nonlinear kernel over the outputs.
        if nonlinear and pi > 0:
            variance = vs.bnd(name='{}/output/nonlin/var'.format(pi), init=1.)
            scales = vs.bnd(name='{}/output/nonlin/scales'.format(pi),
                            init=_vector_from_init(nonlinear_scale, p_num))
            if rq:
                k = RQ(vs.bnd(name='{}/output/nonlin/alpha'.format(pi),
                              init=1e-2, lower=1e-3, upper=1e3))
            else:
                k = EQ()
            kernel_outputs += variance * k.stretch(scales)

        # Construct noise kernel.
        variance = vs.bnd(name='{}/noise'.format(pi),
                          init=_vector_from_init(noise, pi + 1)[pi],
                          lower=1e-8)  # Allow noise to be small.
        kernel_noise = variance * Delta()

        # Construct model and return.
        graph = Graph()
        f = GP(kernel_inputs.select(m_inds) +
               kernel_outputs.select(p_inds), graph=graph)
        e = GP(kernel_noise, graph=graph)
        return f, e

    return model


def _construct_gpar(reg, vs, m, p):
    # Construct GPAR model layer by layer.
    gpar = GPAR(replace=reg.replace, impute=reg.impute, x_ind=reg.x_ind)
    for pi in range(p):
        gpar = gpar.add_layer(_model_generator(vs, m, pi, **reg.model_config))
    return gpar


class GPARRegressor:
    """GPAR regressor.

    Args:
        replace (bool, optional): Replace observations with predictive means.
            Helps the model deal with noisy data points. Defaults to `False`.
        impute (bool, optional): Impute data with predictive means to make the
            data set closed downwards. Helps the model deal with missing data.
            Defaults to `True`.
        scale (tensor, optional): Initial value(s) for the length scale(s) over
            the inputs. Defaults to `1.0`.
        scale_tie (bool, optional): Tie the length scale(s) over the inputs.
            Defaults to `False`.
        per (bool, optional): Use a locally periodic kernel over the inputs.
            Defaults to `False`.
        per_period (tensor, optional): Initial value(s) for the period(s) of the
            locally periodic kernel. Defaults to `1.0`.
        per_scale (tensor, optional): Initial value(s) for the length scale(s)
            of the locally periodic kernel. Defaults to `1.0`.
        per_decay (tensor, optional): Initial value(s) for the length scale(s)
            of the local change of the locally periodic kernel. Defaults to
            `10.0`.
        input_linear (bool, optional): Use a linear kernel over the inputs.
            Defaults to `False`.
        input_linear_scale (tensor, optional): Initial value(s) for the length
            scale(s) of the linear kernel over the inputs. Defaults to `100.0`.
        linear (bool, optional): Use linear dependencies between outputs.
            Defaults to `True`.
        linear_scale (tensor, optional): Initial value(s) for the length
            scale(s) of the linear dependencies. Defaults to `100.0`.
        nonlinear (bool, optional): Use nonlinear dependencies between outputs.
            Defaults to `True`.
        nonlinear_scale (tensor, optional): Initial value(s) for the length
            scale(s) over the outputs. Defaults to `0.1`.
        rq (bool, optional): Use rational quadratic (RQ) kernels instead of
            exponentiated quadratic (EQ) kernels. Defaults to `False`.
        markov (int, optional): Markov order of conditionals. Set to `None` to
            have a fully connected structure. Defaults to `None`.
        noise (tensor, optional): Initial value(s) for the observation noise(s).
            Defaults to `0.01`.
        x_ind (tensor, optional): Locations of inducing points. Set to `None`
            if inducing points should not be used. Defaults to `None`.
        normalise_y (bool, optional): Normalise outputs. Defaults to `True`.
        transform_y (tuple, optional): Tuple containing a transform and its
            inverse, which should be applied to the data before fitting.
            Defaults to the identity transform.

    Attributes:
        replace (bool): Replace observations with predictive means.
        impute (bool): Impute missing data with predictive means to make the
            data set closed downwards.
        sparse (bool): Use inducing points.
        x_ind (tensor): Locations of inducing points.
        model_config (dict): Summary of model configuration.
        vs (:class:`varz.Vars`): Model parameters.
        is_fit (bool): The model is fit.
        x (tensor): Inputs of training data.
        y (tensor): Outputs of training data.
        n (int): Number of training data points.
        m (int): Number of input features.
        p (int): Number of outputs.
        normalise_y (bool): Normalise outputs.
    """

    def __init__(self,
                 replace=False,
                 impute=True,
                 scale=1.0,
                 scale_tie=False,
                 per=False,
                 per_period=1.0,
                 per_scale=1.0,
                 per_decay=10.0,
                 input_linear=False,
                 input_linear_scale=100.0,
                 linear=True,
                 linear_scale=100.0,
                 nonlinear=False,
                 nonlinear_scale=1.0,
                 rq=False,
                 markov=None,
                 noise=0.1,
                 x_ind=None,
                 normalise_y=True,
                 transform_y=(lambda x: x, lambda x: x)):
        # Model configuration.
        self.replace = replace
        self.impute = impute
        self.sparse = x_ind is not None
        if x_ind is None:
            self.x_ind = None
        else:
            self.x_ind = torch.tensor(B.uprank(x_ind))
        self.model_config = {
            'scale': scale,
            'scale_tie': scale_tie,
            'per': per,
            'per_period': per_period,
            'per_scale': per_scale,
            'per_decay': per_decay,
            'input_linear': input_linear,
            'input_linear_scale': input_linear_scale,
            'linear': linear,
            'linear_scale': linear_scale,
            'nonlinear': nonlinear,
            'nonlinear_scale': nonlinear_scale,
            'rq': rq,
            'markov': markov,
            'noise': noise
        }

        # Model fitting.
        self.vs = Vars(dtype=torch.float64)
        self.is_fit = False
        self.x = None  # Inputs of training data
        self.y = None  # Outputs of training data
        self.n = None  # Number of data points
        self.m = None  # Number of input features
        self.p = None  # Number of outputs

        # Output normalisation and transformation.
        self.normalise_y = normalise_y
        self._unnormalise_y, self._normalise_y = lambda x: x, lambda x: x
        self._transform_y, self._untransform_y = transform_y

    def get_variables(self):
        """Construct a dictionary containing all the hyperparameters.

        Returns:
            dict: Dictionary mapping variable names to variable values.
        """
        variables = {}
        for name in self.vs.names:
            variables[name] = self.vs[name].detach().numpy()
        return variables

    def fit(self,
            x,
            y,
            greedy=False,
            fix=True,
            **kw_args):
        """Fit the model to data.

        Further takes in keyword arguments for `Varz.minimise_l_bfgs_b`.

        Args:
            x (tensor): Inputs of training data.
            y (tensor): Outputs of training data.
            greedy (bool, optional): Greedily determine the ordering of the
                outputs. Defaults to `False`.
            fix (bool, optional): Fix the parameters of a layer after
                training it. If set to `False`, the likelihood are
                accumulated and all parameters are optimised at every step.
                Defaults to `True`.
        """
        if greedy:
            raise NotImplementedError('Greedy search is not implemented yet.')

        # Store data.
        self.x = torch.tensor(B.uprank(x))
        self.y = torch.tensor(self._transform_y(B.uprank(y)))
        self.n, self.m = self.x.shape
        self.p = self.y.shape[1]

        # Perform normalisation, carefully handling missing values.
        if self.normalise_y:
            means, stds = [], []
            for i in range(self.p):
                # Filter missing observations.
                available = ~B.isnan(self.y[:, i])
                y_i = self.y[available, i]

                # Calculate mean.
                means.append(B.mean(y_i))

                # Calculate std: safely handle the zero case.
                std = B.std(y_i)
                if std > 0:
                    stds.append(std)
                else:
                    stds.append(B.cast(B.dtype(std), 1))

            # Stack into a vector and create normalisers.
            means, stds = B.stack(*means)[None, :], B.stack(*stds)[None, :]

            def normalise_y(y_):
                return (y_ - means) / stds

            def unnormalise_y(y_):
                return y_ * stds + means

            # Save normalisers.
            self._normalise_y = normalise_y
            self._unnormalise_y = unnormalise_y

            # Perform normalisation.
            self.y = normalise_y(self.y)

        # Precompute the results of `per_output`. This can otherwise incur a
        # significant overhead if the number of outputs is large.
        y_cached = {k: list(per_output(self.y, keep=k)) for k in [True, False]}

        # Fit layer by layer.
        #   Note: `_construct_gpar` takes in the *number* of outputs.
        sys.stdout.write('Training conditionals (total: {}):'.format(self.p))
        sys.stdout.flush()
        for pi in range(self.p):
            sys.stdout.write(' {}'.format(pi + 1))
            sys.stdout.flush()

            # If we fix parameters of previous layers, we can precompute the
            # inputs. This speeds up the optimisation massively.
            if fix:
                gpar = _construct_gpar(self, self.vs, self.m, pi + 1)
                fixed_x, fixed_x_ind = gpar.logpdf(self.x, y_cached,
                                                   only_last_layer=True,
                                                   outputs=list(range(pi)),
                                                   return_inputs=True)

            def objective(vs):
                gpar = _construct_gpar(self, vs, self.m, pi + 1)
                # If the parameters of the previous layers are fixed, use the
                # precomputed inputs.
                if fix:
                    return -gpar.logpdf(fixed_x, y_cached,
                                        only_last_layer=True,
                                        outputs=[pi],
                                        x_ind=fixed_x_ind)
                else:
                    return -gpar.logpdf(self.x, y_cached, only_last_layer=False)

            # Determine names to optimise.
            if fix:
                names = ['{}/*'.format(pi)]
            else:
                names = ['{}/*'.format(i) for i in range(pi + 1)]

            # Perform the optimisation.
            minimise_l_bfgs_b(objective, self.vs, names=names, **kw_args)

        # Print newline to end progress bar.
        sys.stdout.write('\n')

        # Store that the model is fit.
        self.is_fit = True

    def logpdf(self,
               x,
               y,
               sample_missing=False,
               posterior=False,
               differentiable=False):
        """Compute the logpdf of observations.

        Args:
            x (tensor): Inputs.
            y (tensor): Outputs.
            sample_missing (bool, optional): Sample missing data to compute an
                unbiased estimate of the pdf, *not* logpdf. Defaults to `False`.
            posterior (bool, optional): Compute logpdf under the posterior
                instead of the prior. Defaults to `False`.
            differentiable (bool, optional): Do cast the inputs to PyTorch
                tensors and detach the results afterwards. Defaults to `False`.

        Returns
            float: Estimate of the logpdf.
        """
        x = B.uprank(x)
        y = B.uprank(y)

        if not differentiable:
            x = torch.tensor(x)
            y = torch.tensor(y)

        x = B.uprank(x)
        y = self._unnormalise_y(self._transform_y(B.uprank(y)))
        m, p = x.shape[1], y.shape[1]

        if posterior and not self.is_fit:
            raise RuntimeError('Must fit model before computing the logpdf '
                               'under the posterior.')

        # Construct GPAR and compute logpdf.
        gpar = _construct_gpar(self, self.vs, m, p)
        if posterior:
            gpar = gpar | (self.x, self.y)
        value = gpar.logpdf(x, y,
                            only_last_layer=False,
                            sample_missing=sample_missing)

        if not differentiable:
            value = value.detach_().numpy()

        return value

    def sample(self, x, p=None, posterior=False, num_samples=1, latent=False):
        """Sample from the prior or posterior.

        Args:
            x (tensor): Inputs to sample at.
            p (int, optional): Number of outputs to sample if sampling from
                the prior.
            posterior (bool, optional): Sample from the prior instead of the
                posterior.
            num_samples (int, optional): Number of samples. Defaults to `1`.
            latent (bool, optional): Sample the latent function instead of
                observations. Defaults to `False`.

        Returns:
            list[tensor]: Prior samples. If only a single sample is
                generated, it will be returned directly instead of in a list.
        """
        x = torch.tensor(B.uprank(x))

        # Check that model is fit if sampling from the posterior.
        if posterior and not self.is_fit:
            raise RuntimeError('Must fit model before sampling form the '
                               'posterior.')
        # Check that the number of outputs is specified if sampling from the
        # prior.
        elif not posterior and p is None:
            raise ValueError('Must specify number of outputs to sample.')

        if posterior:
            # Construct posterior GPAR.
            gpar = _construct_gpar(self, self.vs, self.m, self.p)
            gpar = gpar | (self.x, self.y)
        else:
            # Construct prior GPAR.
            gpar = _construct_gpar(self, self.vs, B.shape(x)[1], p)

        # Construct function to undo normalisation and transformation.
        def undo_transforms(y_):
            return self._untransform_y(self._unnormalise_y(y_))

        # Perform sampling.
        samples = []
        sys.stdout.write('Sampling (total: {}):'.format(num_samples))
        sys.stdout.flush()
        for i in range(num_samples):
            sys.stdout.write(' {}'.format(i + 1))
            sys.stdout.flush()
            samples.append(undo_transforms(gpar.sample(x, latent=latent))
                           .detach_().numpy())
        sys.stdout.write('\n')
        return samples[0] if num_samples == 1 else samples

    def predict(self, x, num_samples=100, latent=False, credible_bounds=False):
        """Predict at new inputs.

        Args:
            x (tensor): Inputs to predict at.
            num_samples (int, optional): Number of samples. Defaults to `100`.
            latent (bool, optional): Predict the latent function instead of
                observations. Defaults to `True`.
            credible_bounds (bool, optional): Also return 95% central marginal
                credible bounds for the predictions.

        Returns:
            tensor: Predictive means. If `credible_bounds` is set to true,
                a three-tuple will be returned containing the predictive means,
                lower credible bounds, and upper credible bounds.
        """
        # Sample from posterior.
        samples = self.sample(x,
                              num_samples=num_samples,
                              latent=latent,
                              posterior=True)

        # Compute mean.
        mean = np.mean(samples, axis=0)

        if credible_bounds:
            # Also return lower and upper credible bounds if asked for.
            lowers = np.percentile(samples, 2.5, axis=0)
            uppers = np.percentile(samples, 100 - 2.5, axis=0)
            return mean, lowers, uppers
        else:
            return mean



import logging

from lab import B
from stheno import GP, Obs, SparseObs

__all__ = ['GPAR']

log = logging.getLogger(__name__)


def merge(x, updates, to_update):
    """Merge updates into a tensor.

    Args:
        x (tensor): Tensor to merge updates into.
        updates (tensor): Updates.
        to_update (tensor): A boolean array indicating which elements of `x`
            to update and replace with the corresponding element in `updates`.

    Returns:
        tensor: Updated tensor.
    """
    # Stack them, which screws up the order.
    concat = B.concat(x[~to_update], updates, axis=0)

    # Generate an index mapping to fix the ordering.
    i_original = 0
    i_update = B.sum(~to_update)
    indices = []
    for i in range(len(to_update)):
        # Careful not to update the indices in-place! That generates trouble
        # with PyTorch.
        if to_update[i]:
            indices.append(i_update)
            i_update = i_update + 1
        else:
            indices.append(i_original)
            i_original = i_original + 1

    # Perform the fix.
    return B.take(B.dense(concat), indices)


def construct_model(f, e):
    """Convenience function that returns a model constructor.

    Args:
        f (:class:`stheno.GP`): Latent process.
        e (:class:`stheno.GP`): Noise process.

    Returns:
        function: Model constructor.
    """
    return lambda: (f, e)


def last(xs, select=None):
    """Zip a list with a boolean indicating whether it is the last element.

    Args:
        xs (list): List to zip.
        select (list[int], optional): Select particular indices from the result.
            Defaults to returning everything.

    Returns:
        list: `xs` zipped with a boolean on the left.
    """
    # Convert `select` to a set if it is given.
    if select is not None:
        select = set(select)

    # Initialise tracking variables.
    saved_x = None
    i = -1

    # Construct a function that determines the element with index `i` should be
    # yielded.
    def should_yield(i_):
        return i >= 0 and (select is None or i_ in select)

    # Loop through `xs` and check to yield every *previous* element.
    for x in xs:
        if should_yield(i):
            yield False, saved_x
        saved_x = x
        i += 1

    # Check to yield the last element, if there is one.
    if saved_x is not None and should_yield(i):
        yield True, saved_x


class GPAR:
    """Basic GPAR model.

    Args:
        replace (bool, optional): Condition on the predictive mean instead of
            the data. Defaults to `False`.
        impute (bool, optional): Impute missing data points with the predictive
            mean to make the data set closed downwards. Defaults to `False`.
        x_ind (tensor, optional): Locations of inducing points
            for a sparse approximation. Defaults to `None`.
    """

    def __init__(self, replace=False, impute=False, x_ind=None):
        self.replace = replace
        self.impute = impute
        self.layers = []

        # Parse inputs of inducing points.
        self.sparse = x_ind is not None
        self.x_ind = None if x_ind is None else x_ind

    def copy(self):
        """Create a new GPAR model with the same configuration.

        Returns:
            :class:`.gpar.GPAR`: New GPAR model with the same configuration.
        """
        gpar = GPAR(replace=self.replace,
                    impute=self.impute,
                    x_ind=self.x_ind)
        return gpar

    def add_layer(self, model_constructor):
        """Add a layer.

        Args:
            model_constructor (function): Constructor of the model, which should
                return a tuple containing the latent and noise process.

        Returns:
            :class:`.gpar.GPAR`: Updated GPAR model.
        """
        gpar = self.copy()
        gpar.layers = list(self.layers) + [model_constructor]
        return gpar

    def __or__(self, x_and_y):
        """Condition on data.

        Args:
            x (tensor): Inputs.
            y (tensor): Outputs.

        Returns:
            :class:`.gpar.GPAR`: Updated GPAR model.
        """
        x, y = x_and_y
        gpar, x_ind = self.copy(), self.x_ind

        for is_last, ((y, mask), model) in \
                last(zip(per_output(y, self.impute), self.layers)):
            x = x[mask]  # Filter according to mask.
            f, e = model()  # Construct model.
            obs = self._obs(x, x_ind, y, f, e)  # Construct observations.

            # Update with posterior.
            f_post = f | obs
            e_new = GP(e.kernel, e.mean, graph=f.graph)
            gpar.layers.append(construct_model(f_post, e_new))

            # Update inputs.
            if not is_last:
                x, x_ind = self._update_inputs(x, x_ind, y, f, obs)

        return gpar

    def logpdf(self,
               x,
               y,
               only_last_layer=False,
               sample_missing=False,
               return_inputs=False,
               x_ind=None,
               outputs=None):
        """Compute the logpdf.

        Args:
            x (tensor): Inputs.
            y (tensor): Outputs.
            only_last_layer (bool, optional): Compute the logpdf for only the
                last layer. Defaults to `False`.
            sample_missing (bool, optional): Sample missing data to compute an
                unbiased estimate of the pdf, *not* logpdf. Defaults to `False`.
            return_inputs (bool, optional): Instead return the inputs and
                inputs for the inducing points with previous outputs
                concatenated. This can be used to perform precomputation.
                Defaults to `False`.
            x_ind (tensor, optional): Inputs for the inducing points. This
                can be used to resume a computation. Defaults to
                :attr:`.model.GPAR.x_ind`.
            outputs (list[int], optional): Only compute the logpdf for a
                subset of outputs. The list specifies the indices of the
                outputs. Defaults to computing the logpdf for all outputs.

        Returns:
            scalar: Logpdf. If `return_inputs` is set to `True`, instead
                return a tuple containing the inputs and the inputs for the
                inducing points with previous outputs concatenated
        """
        logpdf = B.cast(B.dtype(x), 0)
        x_ind = self.x_ind if x_ind is None else x_ind

        y_per_output = per_output(y, self.impute or sample_missing)
        for is_last, ((y, mask), model) in \
                last(zip(y_per_output, self.layers), select=outputs):
            x = x[mask]  # Filter according to mask.
            f, e = model()  # Construct model.
            obs = self._obs(x, x_ind, y, f, e)  # Construct observations.

            # Accumulate logpdf.
            if not only_last_layer or (is_last and only_last_layer):
                logpdf += f.graph.logpdf(obs)

            if not is_last:
                missing = B.isnan(y[:, 0])
                # Sample missing data for an unbiased sample of the pdf.
                if sample_missing and B.any(missing):
                    y = merge(y, ((f + e) | obs)(x[missing]).sample(), missing)

                # Update inputs.
                x, x_ind = self._update_inputs(x, x_ind, y, f, obs)

        # Return inputs if asked for.
        return (x, x_ind) if return_inputs else logpdf

    def sample(self, x, latent=False):
        """Sample.

        Args:
            x (tensor): Inputs to sample at.
            latent (bool, optional): Sample latent function. Defaults to
                `False`.

        Returns:
            tensor: Sample.
        """
        sample = B.zeros(B.dtype(x), B.shape(x)[0], 0)
        x_ind = self.x_ind

        for is_last, model in last(self.layers):
            f, e = model()  # Construct model.

            if latent:
                # Sample latent function: use ancestral sampling.
                f_sample = f(x).sample()
                y_sample = f_sample + e(x).sample()
                sample = B.concat(sample, f_sample, axis=1)
            else:
                # Sample observed function.
                y_sample = (f + e)(x).sample()
                sample = B.concat(sample, y_sample, axis=1)

            # Update inputs.
            if not is_last:
                x, x_ind = self._update_inputs(x, x_ind, y_sample, f, None)

        return sample

    def _obs(self, x, x_ind, y, f, e):
        available = ~B.isnan(y[:, 0])
        if self.sparse:
            return SparseObs(f(x_ind), e, f(x[available]), y[available])
        else:
            return Obs((f + e)(x[available]), y[available])

    def _update_inputs(self, x, x_ind, y, f, obs):
        available = ~B.isnan(y[:, 0])

        def estimate(x_):
            # If observations are available, estimate using the posterior mean;
            # otherwise, use the prior mean.
            return ((f | obs) if obs else f).mean(x_)

        # Update inputs of inducing points.
        if self.sparse:
            x_ind = B.concat(x_ind, estimate(x_ind), axis=1)

        # Impute missing data and replace available data.
        if self.impute and self.replace:
            y = estimate(x)
        else:
            # Just impute missing data.
            if self.impute and B.any(~available):
                y = merge(y, estimate(x[~available]), ~available)

            # Just replace available data.
            if self.replace and B.any(available):
                y = merge(y, estimate(x[available]), available)

        # Finally, actually update inputs.
        x = B.concat(x, y, axis=1)

        return x, x_ind


def per_output(y, keep=False):
    """Return observations per output, respecting that the data must be
    closed downwards.

    The function supports caching by feeding it a dictionary where the keys
    the are values for `keep` and the values lists containing items
    that the function should yield.

    Args:
        y (tensor): Outputs.
        keep (bool, optional): Also return missing observations that would
            make the data closed downwards.

    Returns:
        generator: Generator that generates tuples containing the
            observations per layer and a mask which observations are not
            missing relative to the previous layer.
    """
    # Handle cache, if that is given.
    if isinstance(y, dict):
        for yi in y[keep]:
            yield yi
        return

    p = B.shape(y)[1]  # Number of outputs
    available = ~B.isnan(y)  # Availability of outputs.

    for i in range(p):
        # Initialise the mask to current availability.
        mask = available[:, i]

        # Take into account future observations if necessary.
        if keep and i < p - 1:  # Check whether this is the last output.
            mask = mask | B.any(available[:, i + 1:], axis=1)

        # Give stuff back.
        yield y[mask, i:i + 1], mask

        # Filter observations and availability.
        y = y[mask]
        available = available[mask]



import numpy as np
import pytest
import torch
from lab.torch import B
from stheno import GP

from gpar.regression import (
    _vector_from_init,
    log_transform,
    squishing_transform,
    GPARRegressor,
    _construct_gpar,
    _determine_indices
)
from .util import allclose, approx, tensor


def test_transforms():
    f, f_inv = log_transform
    allclose(f(f_inv(tensor([1, 2, 3, 4]))), tensor([1, 2, 3, 4]))
    f, f_inv = squishing_transform
    allclose(f(f_inv(tensor([-2, -1, 3, 4]))), tensor([-2, -1, 3, 4]))


def test_vector_from_init():
    allclose(_vector_from_init(2, 2), np.array([2, 2]))
    allclose(_vector_from_init(np.array([1, 2, 3]), 2), np.array([1, 2]))
    with pytest.raises(ValueError):
        _vector_from_init(np.random.randn(2, 2), 1)
    with pytest.raises(ValueError):
        _vector_from_init(np.array([1, 2]), 3)


def test_determine_indices():
    # No Markov structure.
    assert _determine_indices(1, 0, None) == ([0], [], 0)
    assert _determine_indices(1, 1, None) == ([0], [1], 1)
    assert _determine_indices(1, 2, None) == ([0], [1, 2], 2)
    assert _determine_indices(2, 0, None) == ([0, 1], [], 0)
    assert _determine_indices(2, 1, None) == ([0, 1], [2], 1)
    assert _determine_indices(2, 2, None) == ([0, 1], [2, 3], 2)

    # Markov order: 0.
    assert _determine_indices(1, 0, 0) == ([0], [], 0)
    assert _determine_indices(1, 1, 0) == ([0], [], 0)
    assert _determine_indices(1, 2, 0) == ([0], [], 0)
    assert _determine_indices(2, 0, 0) == ([0, 1], [], 0)
    assert _determine_indices(2, 1, 0) == ([0, 1], [], 0)
    assert _determine_indices(2, 2, 0) == ([0, 1], [], 0)

    # Markov order: 1.
    assert _determine_indices(1, 0, 1) == ([0], [], 0)
    assert _determine_indices(1, 1, 1) == ([0], [1], 1)
    assert _determine_indices(1, 2, 1) == ([0], [2], 1)
    assert _determine_indices(2, 0, 1) == ([0, 1], [], 0)
    assert _determine_indices(2, 1, 1) == ([0, 1], [2], 1)
    assert _determine_indices(2, 2, 1) == ([0, 1], [3], 1)

    # Markov order: 2.
    assert _determine_indices(1, 0, 2) == ([0], [], 0)
    assert _determine_indices(1, 1, 2) == ([0], [1], 1)
    assert _determine_indices(1, 2, 2) == ([0], [1, 2], 2)
    assert _determine_indices(2, 0, 2) == ([0, 1], [], 0)
    assert _determine_indices(2, 1, 2) == ([0, 1], [2], 1)
    assert _determine_indices(2, 2, 2) == ([0, 1], [2, 3], 2)


def test_get_variables():
    gpar = GPARRegressor()
    gpar.vs.get(init=1.0, name='variable')
    assert list(gpar.get_variables().items()) == [('variable', 1.0)]


def test_logpdf():
    # Sample some data from a "sensitive" GPAR.
    reg = GPARRegressor(replace=False, impute=False,
                        nonlinear=True, nonlinear_scale=0.1,
                        linear=True, linear_scale=10.,
                        noise=1e-4, normalise_y=False)
    x = np.linspace(0, 5, 10)
    y = reg.sample(x, p=2, latent=True)

    # Extract models.
    gpar = _construct_gpar(reg, reg.vs, 1, 2)
    f1, e1 = gpar.layers[0]()
    f2, e2 = gpar.layers[1]()

    # Test computation under prior.
    logpdf1 = (f1 + e1)(tensor(x)).logpdf(tensor(y[:, 0]))
    x_stack = np.concatenate([x[:, None], y[:, 0:1]], axis=1)
    logpdf2 = (f2 + e2)(tensor(x_stack)).logpdf(tensor(y[:, 1]))
    approx(reg.logpdf(x, y), logpdf1 + logpdf2, digits=6)

    # Test computation under posterior.
    e1_post = GP(e1.kernel, e1.mean, graph=e1.graph)
    e2_post = GP(e2.kernel, e2.mean, graph=e2.graph)
    f1_post = f1 | ((f1 + e1)(tensor(x)), tensor(y[:, 0]))
    f2_post = f2 | ((f2 + e2)(tensor(x_stack)), tensor(y[:, 1]))
    logpdf1 = (f1_post + e1_post)(tensor(x)).logpdf(tensor(y[:, 0]))
    logpdf2 = (f2_post + e2_post)(tensor(x_stack)).logpdf(tensor(y[:, 1]))
    with pytest.raises(RuntimeError):
        reg.logpdf(x, y, posterior=True)
    reg.fit(x, y, iters=0)
    approx(reg.logpdf(x, y, posterior=True), logpdf1 + logpdf2, digits=6)

    # Test that sampling missing gives a stochastic estimate.
    y[::2, 0] = np.nan
    assert np.abs(reg.logpdf(x, y, sample_missing=True) -
                  reg.logpdf(x, y, sample_missing=True)) >= 1e-3


def test_logpdf_differentiable():
    reg = GPARRegressor(replace=False, impute=False,
                        linear=True, linear_scale=1., nonlinear=False,
                        noise=1e-8, normalise_y=False)
    x = np.linspace(0, 5, 10)
    y = reg.sample(x, p=2, latent=True)

    # Test that gradient calculation works.
    reg.vs.requires_grad(True)
    for var in reg.vs.get_vars():
        assert var.grad is None
    reg.logpdf(torch.tensor(x), torch.tensor(y), differentiable=True).backward()
    for var in reg.vs.get_vars():
        assert var.grad is not None


def test_sample_and_predict():
    reg = GPARRegressor(replace=False, impute=False,
                        linear=True, linear_scale=1., nonlinear=False,
                        noise=1e-8, normalise_y=False)
    x = np.linspace(0, 5, 10)

    # Test checks.
    with pytest.raises(ValueError):
        reg.sample(x)
    with pytest.raises(RuntimeError):
        reg.sample(x, posterior=True)

    # Test that output is simplified correctly.
    assert isinstance(reg.sample(x, p=2), np.ndarray)
    assert isinstance(reg.sample(x, p=2, num_samples=2), list)

    # Test that it produces random samples. Not sure how to test correctness.
    assert np.sum(np.abs(reg.sample(x, p=2) - reg.sample(x, p=2))) >= 1e-2
    assert np.sum(np.abs(reg.sample(x, p=2, latent=True) -
                         reg.sample(x, p=2, latent=True))) >= 1e-3

    # Test that mean of posterior samples are around the data.
    y = reg.sample(x, p=2)
    reg.fit(x, y, iters=0)
    approx(y, np.mean(reg.sample(x,
                                 posterior=True,
                                 num_samples=20), axis=0), digits=4)
    approx(y, np.mean(reg.sample(x,
                                 latent=True,
                                 posterior=True,
                                 num_samples=20), axis=0), digits=4)

    # Test that prediction is around the data.
    approx(y, reg.predict(x, num_samples=20), digits=4)
    approx(y, reg.predict(x, latent=True, num_samples=20), digits=4)

    # Test that prediction is confident.
    _, lowers, uppers = reg.predict(x, num_samples=10, credible_bounds=True)
    assert np.less_equal(uppers - lowers, 1e-3).all()


def test_fit():
    reg = GPARRegressor(replace=False, impute=False,
                        normalise_y=True, transform_y=squishing_transform)
    x = np.linspace(0, 5, 10)
    y = reg.sample(x, p=2)

    # TODO: Remove this once greedy search is implemented.
    with pytest.raises(NotImplementedError):
        reg.fit(x, y, greedy=True)

    # Test that data is correctly transformed if it has an output with zero
    # variance.
    reg.fit(x, y, iters=0)
    assert (~B.isnan(reg.y)).numpy().all()
    y_pathological = y.copy()
    y_pathological[:, 0] = 1
    reg.fit(x, y_pathological, iters=0)
    assert (~B.isnan(reg.y)).numpy().all()

    # Test transformation and normalisation of outputs.
    z = torch.linspace(-1, 1, 10, dtype=torch.float64)
    z = B.stack(z, 2 * z, axis=1)
    allclose(reg._untransform_y(reg._transform_y(z)), z)
    allclose(reg._unnormalise_y(reg._normalise_y(z)), z)

    # Test that fitting runs without issues.
    vs = reg.vs.copy(detach=True)
    reg.fit(x, y, fix=False)
    reg.vs = vs
    reg.fit(x, y, fix=True)


def test_features():
    # Test that optimisation runs for a full-fledged GPAR.
    reg = GPARRegressor(replace=True, scale=1.0,
                        per=True, per_period=1.0, per_decay=10.0,
                        input_linear=True, input_linear_scale=0.1,
                        linear=True, linear_scale=1.0,
                        nonlinear=True, nonlinear_scale=1.0,
                        rq=True, noise=0.1)
    x = np.stack([np.linspace(0, 10, 20),
                  np.linspace(10, 20, 20)], axis=1)
    y = reg.sample(x, p=2)
    reg.fit(x, y, iters=10)


def test_scale_tying():
    reg = GPARRegressor(scale_tie=True)
    reg.sample(np.linspace(0, 10, 20), p=2)  # Instantiate variables.
    vs = reg.get_variables()
    assert '0/input/scales' in vs
    assert '1/input/scales' not in vs


def test_inducing_points_uprank():
    reg = GPARRegressor(x_ind=np.linspace(0, 10, 20))
    assert reg.x_ind is not None
    assert B.rank(reg.x_ind) == 2



import numpy as np
import torch
from lab.torch import B
from stheno import GP, EQ, Delta, Graph, Obs, SparseObs, ZeroKernel, Linear

from gpar.model import merge, construct_model, last, per_output, GPAR
from .util import allclose, approx, tensor


def test_merge():
    original = torch.tensor([1, 2, 3, 4])
    updates = torch.tensor([5, 6])

    result = merge(original, updates, torch.tensor([True, True, False, False]))
    allclose(result, [5, 6, 3, 4])

    result = merge(original, updates, torch.tensor([True, False, True, False]))
    allclose(result, [5, 2, 6, 4])


def test_construct_model():
    model = construct_model(1, 2)
    assert model() == (1, 2)


def test_last():
    xs = [1, 2, 3, 4]
    assert list(last(xs)) == [(False, 1), (False, 2), (False, 3), (True, 4)]
    assert list(last(xs, [1, 2])) == [(False, 2), (False, 3)]
    assert list(last(xs, [0, 3])) == [(False, 1), (True, 4)]
    assert list(last([])) == []
    assert list(last([], [0, 1])) == []


def test_per_output():
    y = torch.tensor([[1, 2, np.nan, np.nan],
                      [3, np.nan, 4, np.nan],
                      [5, 6, 7, np.nan],
                      [8, np.nan, np.nan, np.nan],
                      [9, 10, np.nan, np.nan],
                      [11, np.nan, np.nan, 12]])

    expected = [([1, 3, 5, 8, 9, 11], [True, True, True, True, True, True]),
                ([2, 6, 10], [True, False, True, False, True, False]),
                ([7], [False, True, False]),
                ([], [False])]
    result = [(a.numpy()[:, 0].tolist(), b.numpy().tolist())
              for a, b in per_output(y, keep=False)]
    assert result == expected

    expected = [([1, 3, 5, 8, 9, 11], [True, True, True, True, True, True]),
                ([2, -1, 6, 10, -1], [True, True, True, False, True, True]),
                ([4, 7, -1], [False, True, True, False, True]),
                ([12], [False, False, True])]
    result = [([-1 if np.isnan(c) else c for c in a.numpy()[:, 0].tolist()],
               b.numpy().tolist())
              for a, b in per_output(y, keep=True)]
    assert result == expected

    # Test caching.
    assert list(per_output({True: [2, 3], False: [3, 4]}, keep=True)) == [2, 3]
    assert list(per_output({True: [2, 3], False: [4]}, keep=False)) == [4]


def test_misc():
    gpar = GPAR(x_ind=None)
    assert gpar.sparse == False
    assert gpar.x_ind is None

    gpar = GPAR(x_ind=1)
    assert gpar.sparse == True
    assert gpar.x_ind == 1


def test_obs():
    graph = Graph()
    f = GP(EQ(), graph=graph)
    e = GP(1e-8 * Delta(), graph=graph)

    # Check that it produces the correct observations.
    x = torch.linspace(0, 0.1, 10, dtype=torch.float64)
    y = f(x).sample()

    # Set some observations to be missing.
    y_missing = y.clone()
    y_missing[::2] = np.nan

    # Check dense case.
    gpar = GPAR()
    obs = gpar._obs(x, None, y_missing, f, e)
    assert type(obs) == Obs
    approx(y, (f | obs).mean(x))

    # Check sparse case.
    gpar = GPAR(x_ind=x)
    obs = gpar._obs(x, x, y_missing, f, e)
    assert type(obs) == SparseObs
    approx(y, (f | obs).mean(x))


def test_update_inputs():
    graph = Graph()
    f = GP(EQ(), graph=graph)

    x = tensor([[1], [2], [3]])
    y = tensor([[4], [5], [6]])
    res = B.concat(x, y, axis=1)
    x_ind = tensor([[6], [7]])
    res_ind = tensor([[6, 0], [7, 0]])

    # Check vanilla case.
    gpar = GPAR(x_ind=x_ind)
    allclose(gpar._update_inputs(x, x_ind, y, f, None), (res, res_ind))

    # Check imputation with prior.
    gpar = GPAR(impute=True, x_ind=x_ind)
    this_y = y.clone()
    this_y[1] = np.nan
    this_res = res.clone()
    this_res[1, 1] = 0
    allclose(gpar._update_inputs(x, x_ind, this_y, f, None),
             (this_res, res_ind))

    # Check replacing with prior.
    gpar = GPAR(replace=True, x_ind=x_ind)
    this_y = y.clone()
    this_y[1] = np.nan
    this_res = res.clone()
    this_res[0, 1] = 0
    this_res[1, 1] = np.nan
    this_res[2, 1] = 0
    allclose(gpar._update_inputs(x, x_ind, this_y, f, None),
             (this_res, res_ind))

    # Check imputation and replacing with prior.
    gpar = GPAR(impute=True, replace=True, x_ind=x_ind)
    this_res = res.clone()
    this_res[:, 1] = 0
    allclose(gpar._update_inputs(x, x_ind, y, f, None), (this_res, res_ind))

    # Construct observations and update result for inducing points.
    obs = Obs(f(tensor([1, 2, 3, 6, 7])), tensor([9, 10, 11, 12, 13]))
    res_ind = tensor([[6, 12], [7, 13]])

    # Check imputation with posterior.
    gpar = GPAR(impute=True, x_ind=x_ind)
    this_y = y.clone()
    this_y[1] = np.nan
    this_res = res.clone()
    this_res[1, 1] = 10
    allclose(gpar._update_inputs(x, x_ind, this_y, f, obs), (this_res, res_ind))

    # Check replacing with posterior.
    gpar = GPAR(replace=True, x_ind=x_ind)
    this_y = y.clone()
    this_y[1] = np.nan
    this_res = res.clone()
    this_res[0, 1] = 9
    this_res[1, 1] = np.nan
    this_res[2, 1] = 11
    allclose(gpar._update_inputs(x, x_ind, this_y, f, obs), (this_res, res_ind))

    # Check imputation and replacing with posterior.
    gpar = GPAR(impute=True, replace=True, x_ind=x_ind)
    this_res = res.clone()
    this_res[0, 1] = 9
    this_res[1, 1] = 10
    this_res[2, 1] = 11
    allclose(gpar._update_inputs(x, x_ind, y, f, obs), (this_res, res_ind))


def test_conditioning():
    graph = Graph()
    f1, e1 = GP(EQ(), graph=graph), GP(1e-8 * Delta(), graph=graph)
    f2, e2 = GP(EQ(), graph=graph), GP(2e-8 * Delta(), graph=graph)
    gpar = GPAR().add_layer(lambda: (f1, e1)).add_layer(lambda: (f2, e2))

    x = tensor([[1], [2], [3]])
    y = tensor([[4, 5],
                [6, 7],
                [8, 9]])
    gpar = gpar | (x, y)

    # Extract posterior processes.
    f1_post, e1_post = gpar.layers[0]()
    f2_post, e2_post = gpar.layers[1]()

    # Test independence of noises.
    assert graph.kernels[f1_post, e1_post] == ZeroKernel()
    assert graph.kernels[f2_post, e2_post] == ZeroKernel()

    # Test form of noises.
    assert e1.mean == e1_post.mean
    assert e1.kernel == e1_post.kernel
    assert e2.mean == e2_post.mean
    assert e2.kernel == e2_post.kernel

    # Test posteriors.
    approx(f1_post.mean(x), y[:, 0:1])
    approx(f2_post.mean(B.concat(x, y[:, 0:1], axis=1)), y[:, 1:2])


def test_logpdf():
    graph = Graph()
    f1, e1 = GP(EQ(), graph=graph), GP(2e-1 * Delta(), graph=graph)
    f2, e2 = GP(Linear(), graph=graph), GP(1e-1 * Delta(), graph=graph)
    gpar = GPAR().add_layer(lambda: (f1, e1)).add_layer(lambda: (f2, e2))

    # Sample some data from GPAR.
    x = torch.linspace(0, 2, 10, dtype=torch.float64)[:, None]
    y = gpar.sample(x, latent=True)

    # Compute logpdf.
    logpdf1 = (f1 + e1)(x).logpdf(y[:, 0])
    logpdf2 = (f2 + e2)(B.concat(x, y[:, 0:1], axis=1)).logpdf(y[:, 1])

    # Test computation of GPAR.
    assert gpar.logpdf(x, y) == logpdf1 + logpdf2
    assert gpar.logpdf(x, y, only_last_layer=True) == logpdf2

    # Test resuming computation.
    x_int, x_ind_int = gpar.logpdf(x, y, return_inputs=True, outputs=[0])
    assert gpar.logpdf(x_int, y, x_ind=x_ind_int, outputs=[1]) == logpdf2

    # Test that sampling missing gives a stochastic estimate.
    y[1, 0] = np.nan
    assert B.abs(gpar.logpdf(x, y, sample_missing=True) -
                 gpar.logpdf(x, y, sample_missing=True)).numpy() >= 1e-3


def test_sample():
    graph = Graph()
    x = tensor([1, 2, 3])[:, None]

    # Test that it produces random samples. Not sure how to test for
    # correctness.
    f1, e1 = GP(EQ(), graph=graph), GP(1e-1 * Delta(), graph=graph)
    f2, e2 = GP(EQ(), graph=graph), GP(1e-1 * Delta(), graph=graph)
    gpar = GPAR().add_layer(lambda: (f1, e1)).add_layer(lambda: (f2, e2))
    assert B.sum(B.abs(gpar.sample(x) - gpar.sample(x))) >= 1e-3
    assert B.sum(B.abs(gpar.sample(x, latent=True) -
                       gpar.sample(x, latent=True))) >= 1e-3

    # Test that posterior latent samples are around the data that is
    # conditioned on.
    graph = Graph()
    f1, e1 = GP(EQ(), graph=graph), GP(1e-8 * Delta(), graph=graph)
    f2, e2 = GP(EQ(), graph=graph), GP(1e-8 * Delta(), graph=graph)
    gpar = GPAR().add_layer(lambda: (f1, e1)).add_layer(lambda: (f2, e2))
    y = gpar.sample(x, latent=True)
    gpar = gpar | (x, y)
    approx(gpar.sample(x), y, digits=3)
    approx(gpar.sample(x, latent=True), y, digits=3)



import torch
from numpy.testing import assert_allclose, assert_array_almost_equal
from plum import Dispatcher
from matrix import AbstractMatrix
import lab as B

_dispatch = Dispatcher()


@_dispatch(object)
def convert(a):
    return a


@_dispatch(torch.Tensor)
def convert(a):
    return a.numpy()


@_dispatch(AbstractMatrix)
def convert(a):
    return B.dense(a)


@_dispatch(object, object)
def allclose(a, b):
    assert_allclose(convert(a), convert(b))


@_dispatch(tuple, tuple)
def allclose(a, b):
    if len(a) != len(b):
        raise AssertionError('Inputs "{}" and "{}" are not of the same length.'
                             ''.format(a, b))
    for x, y in zip(a, b):
        assert_allclose(convert(x), convert(y))


@_dispatch(object, object, [object])
def approx(a, b, digits=4):
    assert_array_almost_equal(convert(a), convert(b), decimal=digits)


@_dispatch(tuple, tuple, [object])
def approx(a, b, digits=4):
    if len(a) != len(b):
        raise AssertionError('Inputs "{}" and "{}" are not of the same length.'
                             ''.format(a, b))
    for x, y in zip(a, b):
        approx(x, y, digits=digits)


def tensor(x):
    """Construct a PyTorch tensor of type `torch.float64`.

    Args:
        x (obj): Object to construct array from.

    Returns:
        tensor: PyTorch array of type `torch.float64`.

    """
    return torch.tensor(x, dtype=torch.float64)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from setuptools import find_packages, setup

requirements = ['numpy>=1.16',
                'torch',

                'plum-dispatch',
                'backends',
                'backends-matrix',
                'stheno',
                'varz']

setup(packages=find_packages(exclude=['docs']),
      python_requires='>=3.6',
      install_requires=requirements,
      include_package_data=True)
